require 'forwardable'
require 'multi_git/utils'
require 'multi_git/handle'
require 'multi_git/builder'
require 'multi_git/object'
module MultiGit

  module Commit

    module Base
      extend Utils::AbstractMethods
      extend Forwardable

      # @!attribute message
      #   @return [String]
      abstract :message

      # @!attribute tree
      #   @return [Tree]
      abstract :tree

      # @!attribute parents
      #   @return [Array<Commit>]
      abstract :parents

      # @!attribute time
      #   @return [Time]
      abstract :time

      # @!attribute author
      #   @return [Handle]
      abstract :author

      # @!attribute commit_time
      #   @return [Time]
      abstract :commit_time

      # @!attribute committer
      #   @return [Handle]
      abstract :committer

      # @!method [](path)
      #   @see (MultiGit::Tree#[])
      delegate :[] => :tree

      # @!parse alias / []
      delegate :/ => :tree


      def type
        :commit
      end
    end

    # A commit builder helps creating new commits by 
    # providing a simple interface, sane defaults and some
    # validations.
    #
    # You can create a new commit using the commit builder like this:
    # 
    # @example Commit an small tree
    #   #setup:
    #   dir = `mktemp -d`
    #   # example:
    #   builder = MultiGit::Commit::Builder.new
    #   builder.message = "My first commit"
    #   builder.by "me@example.com"
    #   builder["a_file"] = "some content"
    #   # builder is now ready to be inserted
    #   repository = MultiGit.open(dir, init: true)
    #   commit = repository << builder #=> be_a MultiGit::Commit
    #   commit['a_file'].content  #=> eql "some content"
    #   # teardown:
    #   `rm -rf #{dir}`
    class Builder
      include MultiGit::Builder
      include Base

      # @return (see MultiGit::Commit::Base#message)
      def message(*args)
        if args.any?
          self.message = args.first
          return self
        else
          return @message
        end
      end

      #
      # @yield allows 
      # @return [Tree::Builder]
      def tree(&block)
        @tree.instance_eval(&block) if block
        @tree
      end
      # @return [Array<Commit::Base>]
      attr :parents

      # @return (see MultiGit::Commit::Base#time)
      attr :time
      # @return (see MultiGit::Commit::Base#commit_time)
      attr :commit_time

      # @return (see MultiGit::Commit::Base#author)
      attr :author
      # @return (see MultiGit::Commit::Base#committer)
      attr :committer

      attr_writer :author, :committer
      attr_writer :message

      # @param time [Time]
      def time=(time)
        raise ArgumentError, "Expected a Time, got #{time.inspect}" unless time.kind_of? Time
        @time = time
      end

      # @param time [Time]
      def commit_time=(time)
        raise ArgumentError, "Expected a Time, got #{time.inspect}" unless time.kind_of? Time
        @commit_time = time
      end

      # @param handle [Handle, String]
      def committer=(handle)
        @committer = parse_handle(handle)
      end

      # @param handle [Handle, String]
      def author=(handle)
        @author = parse_handle(handle)
      end

      # DSL method to set author and committer in one step
      # @param handle [Handle, String]
      # @return self
      def by(handle)
        self.author = self.committer = handle
        return self
      end

      def at(time)
        self.time = self.commit_time = time
        return self
      end

      extend Forwardable
      # @!method []=(path, content)
      #   @see (MultiGit::TreeBuilder#[]=)
      delegate :[]= => :tree

      def initialize(from = nil, &block)
        @parents = []
        if from.kind_of? Tree
          @tree = from.to_builder
        elsif from.kind_of? Tree::Builder
          @tree = from
        elsif from.kind_of? Commit
          @tree = from.tree.to_builder
          @parents << from
        elsif from.nil?
          @tree = Tree::Builder.new
        end
        @author = nil
        @committer = nil
        @time = @commit_time = Time.now
        instance_eval(&block) if block
      end

      def >>(repo)
        new_tree = repo << tree
        new_parents = parents.map{|p| repo.write(p).oid }
        c = committer || author    || Handle::DEFAULT
        a = author    || committer || Handle::DEFAULT
        return repo.make_commit(
          :time => time,
          :author => a,
          :commit_time => commit_time,
          :committer => c,
          :parents => new_parents,
          :tree => new_tree.oid,
          :message => message || "Thank you for using multi_git",
          :update_ref => []
        )
      end

    protected

      def parse_handle(handle)
        case(handle)
        when Handle then
          return handle
        when String then
          return Handle.parse(handle)
        else
          raise ArgumentError, "Expected a String or a Handle, got #{handle.inspect}"
        end
      end

    end

    include Object
    include Base

  end
end

